import pdal

"""
Some useful filters combinations for complete pdal pipeline
"""


def add_radius_assign(
    pipeline: pdal.Pipeline,
    radius: float,
    search_3d: bool,
    condition_src: str,
    condition_ref: str,
    condition_out: str,
    max2d_above: float = -1,
    max2d_below: float = -1,
) -> pdal.Pipeline:
    """
    Search points from "condition_src" that are closer than "radius_search" from points that
    belong to "condition_ref" and modify them with "condition_out"

    This combination is equivalent to the CloseBy macro of TerraScan

    Args:
        pipeline (pdal.Pipeline): pdal pipeline
        radius (float): search distance
        search_3d (bool): the distance research is in 3d if True (2d otherwise)
        condition_src (str): pdal condition for points to apply the modification to (eg. "Classification==2")
        condition_ref (str): pdal condition for the potential neighbors to search for (eg. "Classification==4")
        condition_out (str): pdal condition to apply to the points that belong to "condition_src" and
            have a point from "condition_ref" closer than "radius" (eg. "Classification==2")
        max2d_above (float, optional): In case of 2d Search, upward limit for potential neighbors. Defaults to -1.
        max2d_below (float, optional):  In case of 2d Search, downward limit for potential neighbors. Defaults to -1.

    Returns:
        pdal.Pipeline: output pipeline with the radius_assign steps added.
    """

    pipeline |= pdal.Filter.ferry(dimensions="=>REF_DOMAIN, =>SRC_DOMAIN, =>radius_search")
    pipeline |= pdal.Filter.assign(
        value=[
            "SRC_DOMAIN = 0",
            f"SRC_DOMAIN = 1 WHERE {condition_src}",
            "REF_DOMAIN = 0",
            f"REF_DOMAIN = 1 WHERE {condition_ref}",
            "radius_search = 0",
        ]
    )
    pipeline |= pdal.Filter.radius_assign(
        radius=radius,
        src_domain="SRC_DOMAIN",
        reference_domain="REF_DOMAIN",
        output_dimension="radius_search",
        is3d=search_3d,
        max2d_above=max2d_above,
        max2d_below=max2d_below,
    )
    pipeline |= pdal.Filter.assign(value=condition_out, where="radius_search==1")
    return pipeline


def classify_hgt_ground(pipeline, h_min, h_max, condition, condition_out):
    """
    reassign points from "condition" between "h_min" and "h_max" of the ground to "condition_out"
    This combination is equivalent to the ClassifyHgtGrd macro of TerraScan
    condition, condition_out : a pdal condition as "Classification==2"
    """
    pipeline |= pdal.Filter.hag_delaunay(allow_extrapolation=True)
    condition_h = f"HeightAboveGround>{h_min} && HeightAboveGround<={h_max}"
    condition_h += " && " + condition
    pipeline |= pdal.Filter.assign(value=condition_out, where=condition_h)
    return pipeline


def keep_non_planar_pts(pipeline, condition, condition_out):
    """
    reassign points from "condition" who are planar to "condition_out"
    This combination is equivalent to the ClassifyModelKey macro of TerraScan
    condition, condition_out : a pdal condition as "Classification==2"
    """
    pipeline |= pdal.Filter.approximatecoplanar(knn=8, thresh1=25, thresh2=6, where=condition)
    pipeline |= pdal.Filter.assign(value=condition_out, where=f"Coplanar==0 && ({condition})")
    return pipeline


def build_condition(key, values):
    """
    build 'key==values[0] || key==values[1] ...'
    """
    condition = "("
    for v in values:
        condition += key + "==" + str(v)
        if v != values[-1]:
            condition += " || "
    condition += ")"
    return condition



import csv
import json
import math
import tempfile
from test import utils

import numpy as np
import pdal
import pdaltools.las_info as li
import pytest


def contains(bounds, x, y):
    # to be coherent with the grid decimation algorithm
    return bounds[0] <= x and x < bounds[1] and bounds[2] <= y and y < bounds[3]


def run_filter(output_type, resolution):

    ini_las = "test/data/4_6.las"

    tmp_out_wkt = tempfile.NamedTemporaryFile(suffix=f"_{resolution}.wkt").name

    filter_name = "filters.grid_decimation_deprecated"
    utils.pdal_has_plugin(filter_name)

    bounds = li.las_get_xy_bounds(ini_las)

    d_width = math.floor((bounds[0][1] - bounds[0][0]) / resolution) + 1
    d_height = math.floor((bounds[1][1] - bounds[1][0]) / resolution) + 1
    nb_dalle = d_width * d_height

    PIPELINE = [
        {"type": "readers.las", "filename": ini_las},
        {
            "type": filter_name,
            "resolution": resolution,
            "output_type": output_type,
            "output_dimension": "grid",
            "output_wkt": tmp_out_wkt,
        },
    ]

    pipeline = pdal.Pipeline(json.dumps(PIPELINE))

    # execute the pipeline
    pipeline.execute()
    arrays = pipeline.arrays
    array = arrays[0]

    nb_pts_grid = 0
    for pt in array:
        if pt["grid"] > 0:
            nb_pts_grid += 1

    assert nb_pts_grid == nb_dalle

    for lig in range(d_height):
        for col in range(d_width):

            cell = [
                bounds[0][0] + col * resolution,
                bounds[0][0] + (col + 1) * resolution,
                bounds[1][0] + lig * resolution,
                bounds[1][0] + (lig + 1) * resolution,
            ]

            nbThreadPtsCrop = 0
            ZRef = 0
            ZRefGrid = 0

            for pt in array:
                x = pt["X"]
                y = pt["Y"]
                if not contains(cell, x, y):
                    continue

                z = pt["Z"]
                if output_type == "max":
                    if ZRef == 0 or z > ZRef:
                        ZRef = z
                elif output_type == "min":
                    if ZRef == 0 or z < ZRef:
                        ZRef = z

                if pt["grid"] > 0:
                    nbThreadPtsCrop += 1
                    ZRefGrid = z

            assert nbThreadPtsCrop == 1
            assert ZRef == ZRefGrid

    data = []
    with open(tmp_out_wkt, "r") as f:
        reader = csv.reader(f, delimiter="\t")
        for i, line in enumerate(reader):
            data.append(line[0])

    assert len(data) == nb_dalle


@pytest.mark.parametrize(
    "resolution",
    [(10.1), (10.0), (9.8)],
)
def test_grid_decimation_max(resolution):
    run_filter("max", resolution)


@pytest.mark.parametrize(
    "resolution",
    [(10.3), (10.0), (9.9)],
)
def test_grid_decimation_min(resolution):
    run_filter("min", resolution)


def test_grid_decimation_empty():
    ini_las = "test/data/4_6.las"
    with tempfile.NamedTemporaryFile(suffix="_empty.wkt") as tmp_out_wkt:
        pipeline = pdal.Pipeline() | pdal.Reader.las(filename=ini_las)
        pipeline |= pdal.Filter.grid_decimation_deprecated(
            resolution=10,
            output_type="min",
            output_dimension="grid",
            output_wkt=tmp_out_wkt.name,
            where="Classification==123",  # should create an empty result
        )
        pipeline.execute()

        with open(tmp_out_wkt.name, "r") as f:
            reader = csv.reader(f, delimiter="\t")
            lines = [line for line in reader]
            assert len(lines) == 0

        assert np.all(pipeline.arrays[0]["grid"] == 0)



import argparse

import pdal

from pdal_ign_macro import macro

"""
This tool shows how to use functions of macro in a pdal pipeline
"""


def parse_args():
    parser = argparse.ArgumentParser("Tool to apply pdal pipelines for DSM and DTM calculation")
    parser.add_argument("--input", "-i", type=str, required=True, help="Input las file")
    parser.add_argument(
        "--output_las", "-o", type=str, required=True, help="Output cloud las file"
    )
    parser.add_argument("--output_dsm", "-s", type=str, required=True, help="Output dsm tiff file")
    parser.add_argument("--output_dtm", "-t", type=str, required=True, help="Output dtm tiff file")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()

    pipeline = pdal.Reader.las(args.input)

    # 1 - recherche des points max de végétation (4,5) sur une grille régulière, avec prise en compte des points sol (2) et basse
    #     vegetation (3) proche de la végétation : on les affecte en 100

    # bouche trou : assigne les points sol en 102 à l'intérieur de la veget (4,5)
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="Classification==2",
        condition_ref=macro.build_condition("Classification", [4, 5]),
        condition_out="Classification=102",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="Classification==102",
        condition_ref="Classification==2",
        condition_out="Classification=2",
    )

    # selection des points de veget basse proche de la veget haute : assigne 103
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="Classification==3",
        condition_ref="Classification==5",
        condition_out="Classification=103",
    )

    # max des points de veget (et surement veget - 102,103) sur une grille régulière : assigne 100
    pipeline |= pdal.Filter.gridDecimation(
        resolution=0.75,
        value="Classification=100",
        output_type="max",
        where=macro.build_condition("Classification", [4, 5, 102, 103]),
    )

    # remise à zero des codes 102 et 103
    pipeline |= pdal.Filter.assign(value="Classification=2", where="Classification==102")
    pipeline |= pdal.Filter.assign(value="Classification=3", where="Classification==103")

    # 2 - sélection des points pour DTM et DSM

    # selection de points sol (max) sur une grille régulière
    pipeline |= pdal.Filter.gridDecimation(
        resolution=0.5, value="Classification=102", output_type="max", where="Classification==2"
    )

    # selection de points DSM (max) sur une grille régulière
    pipeline |= pdal.Filter.gridDecimation(
        resolution=0.5,
        value="Classification=200",
        output_type="max",
        where=macro.build_condition("Classification", [2, 3, 4, 5, 6, 9, 17, 64, 100]),
    )

    # assigne des points sol sélectionnés (102) en 100 : les points proches de la végaétation, des ponts, de l'eau et 64
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==102",
        condition_ref=macro.build_condition("Classification", [4, 5, 6, 9, 17, 64, 100]),
        condition_out="Classification=100",
    )

    # remise à zero du code 102
    pipeline |= pdal.Filter.assign(value="Classification=2", where="Classification==102")

    # 3 - gestion des ponts

    # bouche trou : on élimine les points sol (2) au milieu du pont en les mettant à 102
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==2",
        condition_ref="Classification==17",
        condition_out="Classification=102",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==102",
        condition_ref=macro.build_condition("Classification", [2, 3, 4, 5]),
        condition_out="Classification=2",
    )

    # bouche trou : on élimine les points basse végétation (3) au milieu du pont en les mettant à 103
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==3",
        condition_ref="Classification==17",
        condition_out="Classification=103",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==103",
        condition_ref=macro.build_condition("Classification", [2, 3, 4, 5]),
        condition_out="Classification=3",
    )

    # bouche trou : on élimine les points moyenne végétation (4) au milieu du pont en les mettant à 104
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==4",
        condition_ref="Classification==17",
        condition_out="Classification=104",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==104",
        condition_ref=macro.build_condition("Classification", [2, 3, 4, 5]),
        condition_out="Classification=4",
    )

    # bouche trou : on élimine les points haute végétation (5) au milieu du pont en les mettant à 105
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==5",
        condition_ref="Classification==17",
        condition_out="Classification=105",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==105",
        condition_ref=macro.build_condition("Classification", [2, 3, 4, 5]),
        condition_out="Classification=5",
    )

    # bouche trou : on élimine les points eau (9) au milieu du pont en les mettant à 109
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==9",
        condition_ref="Classification==17",
        condition_out="Classification=109",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src="Classification==109",
        condition_ref="Classification==9",
        condition_out="Classification=9",
    )

    # step 15 et supression des points ??

    # 4 - export du nuage
    pipeline |= pdal.Writer.las(extra_dims="all", forward="all", filename=args.output_las)

    # export des DSM/DTM
    pipeline |= pdal.Writer.gdal(
        gdaldriver="GTiff",
        output_type="max",
        resolution=2.0,
        filename=args.output_dtm,
        where=macro.build_condition("Classification", [2, 66]),
    )
    pipeline |= pdal.Writer.gdal(
        gdaldriver="GTiff",
        output_type="max",
        resolution=2.0,
        filename=args.output_dsm,
        where=macro.build_condition("Classification", [2, 3, 4, 5, 17, 64]),
    )

    pipeline.execute()



import json
import math
import random as rand
import tempfile
from test import utils

import numpy as np
import pdal
import pytest

pt_x = 1639825.1
pt_y = 1454924.6
pt_z = 7072.1
pt_ini = (pt_x, pt_y, pt_z, 1)

numeric_precision = 4


def distance2d(pt1, pt2):
    return round(math.sqrt((pt1[0] - pt2[0]) ** 2 + (pt1[1] - pt2[1]) ** 2), numeric_precision)


def distance3d(pt1, pt2):
    return round(
        math.sqrt((pt1[0] - pt2[0]) ** 2 + (pt1[1] - pt2[1]) ** 2 + (pt1[2] - pt2[2]) ** 2),
        numeric_precision,
    )


def run_filter(arrays_las, distance_radius, search_3d, limit_z_above=-1, limit_w_below=-1):

    filter = "filters.radius_assign"
    utils.pdal_has_plugin(filter)

    with tempfile.NamedTemporaryFile(suffix="_las_tmp.las") as las:
        pipeline = pdal.Writer.las(filename=las.name).pipeline(arrays_las)
        pipeline.execute()

        PIPELINE = [
            {"type": "readers.las", "filename": las.name},
            {"type": "filters.ferry", "dimensions": "=>SRC_DOMAIN"},
            {"type": "filters.ferry", "dimensions": "=>REF_DOMAIN"},
            {
                "type": "filters.assign",
                "value": [
                    "SRC_DOMAIN = 1 WHERE Classification==2",
                    "SRC_DOMAIN = 0 WHERE Classification!=2",
                    "REF_DOMAIN = 1 WHERE Classification==1",
                    "REF_DOMAIN = 0 WHERE Classification!=1",
                ],
            },
            {
                "type": filter,
                "radius": distance_radius,
                "src_domain": "SRC_DOMAIN",
                "reference_domain": "REF_DOMAIN",
                "output_dimension": "radius_search",
                "is3d": search_3d,
                "max2d_above": limit_z_above,
                "max2d_below": limit_w_below,
            },
        ]

        pipeline = pdal.Pipeline(json.dumps(PIPELINE))
        pipeline.execute()
        arrays = pipeline.arrays
        array = arrays[0]

    nb_pts_radius_search = 0
    for pt in array:
        if pt["radius_search"] > 0:
            nb_pts_radius_search += 1

    return nb_pts_radius_search


def build_random_points_around_one_point(test_function, distance_radius):

    dtype = [("X", "<f8"), ("Y", "<f8"), ("Z", "<f8"), ("Classification", "u1")]
    arrays_las = np.array([pt_ini], dtype=dtype)

    pt_limit = (pt_x + distance_radius, pt_y, pt_z, 2)
    arrays_pti = np.array([pt_limit], dtype=dtype)
    arrays_las = np.concatenate((arrays_las, arrays_pti), axis=0)
    nb_points_take = test_function(pt_limit)

    pt_limit = (pt_x + distance_radius + 1 / numeric_precision, pt_y, pt_z, 2)
    arrays_pti = np.array([pt_limit], dtype=dtype)
    arrays_las = np.concatenate((arrays_las, arrays_pti), axis=0)
    nb_points_take += test_function(pt_limit)

    nb_points = rand.randint(20, 50)
    for i in range(nb_points):
        # round at 1 to avoid precision numeric pb
        pti_x = pt_ini[0] + rand.uniform(-1.5, 1.5)
        pti_y = pt_ini[1] + rand.uniform(-1.5, 1.5)
        pti_z = pt_ini[2] + rand.uniform(-1.5, 1.5)
        pt_i = (pti_x, pti_y, pti_z, 2)

        # too much uncertainty between the digital precisions of pdal and the tests
        if abs(distance2d(pt_i, pt_ini) - distance_radius) < numeric_precision:
            continue
        if abs(distance3d(pt_i, pt_ini) - distance_radius) < numeric_precision:
            continue

        arrays_pti = np.array([pt_i], dtype=dtype)
        arrays_las = np.concatenate((arrays_las, arrays_pti), axis=0)

        nb_points_take += test_function(pt_i)

    return arrays_las, nb_points_take


def test_radius_assign_3d():

    distance_radius = 1

    def func_test(pt):
        distance_i = distance3d(pt_ini, pt)
        if distance_i < distance_radius:
            return 1
        return 0

    arrays_las, nb_points_take_3d = build_random_points_around_one_point(
        func_test, distance_radius
    )
    nb_pts_radius_3d = run_filter(arrays_las, distance_radius, True)
    assert nb_pts_radius_3d == nb_points_take_3d


def test_radius_assign_2d():

    distance_radius = 1

    def func_test(pt):
        distance_i = distance2d(pt_ini, pt)
        if distance_i < distance_radius:
            return 1
        return 0

    arrays_las, nb_points_take_2d = build_random_points_around_one_point(
        func_test, distance_radius
    )
    nb_pts_radius_2d = run_filter(arrays_las, distance_radius, False)
    assert nb_pts_radius_2d == nb_points_take_2d


@pytest.mark.parametrize(
    "limit_z_above, limit_z_below",
    [
        (-1, -1),  # no limit
        (-1, 2),  # limit below only
        (2, -1),  # limit above only
        (0, -1),  # take all points below only
        (-1, 0),  # take all points above only
    ],
)
def test_radius_assign_2d_cylinder(limit_z_above, limit_z_below):

    distance_radius = 1
    limit_z_above = 0.25
    limit_z_below = 0.25

    def func_test(pt):
        distance_i = distance2d(pt_ini, pt)
        if distance_i < distance_radius:
            if (limit_z_above >= 0) and ((pt[2] - pt_ini[2]) <= limit_z_above):
                return 1
            if (limit_z_below >= 0) and ((pt_ini[2] - pt[2]) <= limit_z_below):
                return 1
        return 0

    arrays_las, nb_points_take_2d = build_random_points_around_one_point(
        func_test, distance_radius
    )
    nb_pts_radius_2d_cylinder = run_filter(
        arrays_las, distance_radius, False, limit_z_above, limit_z_below
    )
    assert nb_pts_radius_2d_cylinder == nb_points_take_2d



import argparse
import shutil
import tempfile

import pdal
from pdaltools.las_add_buffer import run_on_buffered_las
from pdaltools.las_remove_dimensions import remove_dimensions_from_las

from pdal_ign_macro import macro

"""
This tool applies a pdal pipeline to select points for DSM and DTM calculation
It adds dimensions with positive values for the selected points
"""


def parse_args():
    parser = argparse.ArgumentParser(
        "Tool to apply pdal pipelines to select points for DSM and DTM calculation"
        + "(add dimensions with positive values for the selected points)"
    )
    parser.add_argument("--input_las", "-i", type=str, required=True, help="Input las file")
    parser.add_argument(
        "--output_las", "-o", type=str, required=True, help="Output cloud las file"
    )
    parser.add_argument(
        "--dsm_dimension",
        type=str,
        required=False,
        default="dsm_marker",
        help="Dimension name for the output DSM marker",
    )
    parser.add_argument(
        "--dtm_dimension",
        type=str,
        required=False,
        default="dtm_marker",
        help="Dimension name for the output DTM marker",
    )
    parser.add_argument(
        "--output_dsm", "-s", type=str, required=False, default="", help="Output dsm tiff file"
    )
    parser.add_argument(
        "--output_dtm", "-t", type=str, required=False, default="", help="Output dtm tiff file"
    )
    parser.add_argument(
        "--keep_temporary_dims",
        "-k",
        action="store_true",
        help="If set, do not delete temporary dimensions",
    )
    parser.add_argument(
        "--skip_buffer",
        "-s",
        action="store_true",
        help="If set, skip adding a buffer from the neighbor tiles based on their name",
    )
    parser.add_argument(
        "--buffer_width",
        type=float,
        default=25,
        help="width of the border to add to the tile (in meters)",
    )
    parser.add_argument(
        "--spatial_ref",
        type=str,
        default="EPSG:2154",
        help="spatial reference for the writer (required when running with a buffer)",
    )
    parser.add_argument(
        "--tile_width",
        type=int,
        default=1000,
        action="store_true",
        help="width of tiles in meters (required when running with a buffer)",
    )
    parser.add_argument(
        "--tile_coord_scale",
        type=int,
        default=1000,
        action="store_true",
        help="scale used in the filename to describe coordinates in meters (required when running with a buffer)",
    )

    return parser.parse_args()


def define_marking_pipeline(input_las, output_las, dsm_dimension, dtm_dimension):
    pipeline = pdal.Pipeline() | pdal.Reader.las(input_las)

    # 0 - ajout de dimensions temporaires et de sortie
    temporary_dimensions = [
        "PT_VEG_DSM",
        "PT_ON_BRIDGE",
        "PT_ON_BUILDING",
        "PT_ON_VEGET",
        "PT_ON_SOL",
    ]
    added_dimensions = [dtm_dimension, dsm_dimension] + temporary_dimensions

    pipeline |= pdal.Filter.ferry(dimensions="=>" + ", =>".join(added_dimensions))

    # 1 - recherche des points max de végétation (4,5) sur une grille régulière, avec prise en
    # compte des points sol (2) et basse
    #     vegetation (3) proches de la végétation
    #     pour le calcul du DSM

    pipeline |= pdal.Filter.assign(
        value=["PT_VEG_DSM = 1 WHERE " + macro.build_condition("Classification", [4, 5])]
    )

    # bouche trou : assigne les points sol à l'intérieur de la veget (4,5)
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="Classification==2",
        condition_ref=macro.build_condition("Classification", [4, 5]),
        condition_out="PT_VEG_DSM=1",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src=macro.build_condition("Classification", [6, 9, 17, 67]),
        condition_ref=macro.build_condition("Classification", [4, 5]),
        condition_out="PT_ON_VEGET=1",
        max2d_above=0,  # ne pas prendre les points qui sont au dessus des points pont (condition_ref)
        max2d_below=-1,  # prendre tous les points qui sont en dessous des points pont (condition_ref)
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="PT_VEG_DSM==1 && Classification==2",
        condition_ref="Classification==2 && PT_VEG_DSM==0",
        condition_out="PT_VEG_DSM=0",
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="PT_ON_VEGET==1 && " + macro.build_condition("Classification", [6, 67]),
        condition_ref="PT_ON_VEGET==0 &&" + macro.build_condition("Classification", [6, 67]),
        condition_out="PT_ON_VEGET=0",
        max2d_above=0.5,  # ne pas  prendre les points qui sont au dessus des points pont (condition_ref)
        max2d_below=0.5,  # prendre tous les points qui sont en dessous des points pont (condition_ref)
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="PT_ON_VEGET==1 && Classification==17",
        condition_ref="Classification==17 && PT_ON_VEGET==0",
        condition_out="PT_ON_VEGET=0",
        max2d_above=0.5,  # ne pas  prendre les points qui sont au dessus des points pont (condition_ref)
        max2d_below=0.5,  # prendre tous les points qui sont en dessous des points pont (condition_ref)
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="PT_ON_VEGET==1 && Classification==9",
        condition_ref="Classification==9 && PT_ON_VEGET==0",
        condition_out="PT_ON_VEGET=0",
        max2d_above=0.5,  # ne pas  prendre les points qui sont au dessus des points pont (condition_ref)
        max2d_below=0.5,  # prendre tous les points qui sont en dessous des points pont (condition_ref)
    )

    # selection des points de veget basse proche de la veget haute
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="Classification==3",
        condition_ref="Classification==5",
        condition_out="PT_VEG_DSM=1",
    )

    # max des points de veget (PT_VEG_DSM==1) sur une grille régulière :
    # TODO: remplacer par GridDecimation une fois le correctif mergé dans PDAL
    pipeline |= pdal.Filter.grid_decimation_deprecated(
        resolution=0.75, output_dimension=dsm_dimension, output_type="max", where="PT_VEG_DSM==1"
    )

    # 2 - sélection des points pour DTM et DSM

    ###########################################################################################################################################
    # Voir pour l'eau sous le sol dans les dévers corse
    pipeline = macro.add_radius_assign(
        pipeline,
        1.25,
        False,
        condition_src="Classification==9",
        condition_ref="Classification==2",
        condition_out="PT_ON_SOL=1",
        max2d_above=0,
        max2d_below=-1,
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src="PT_ON_SOL==1",
        condition_ref="PT_ON_SOL==0 && Classification==9",
        condition_out="PT_ON_SOL=0",
        max2d_above=0.5,
        max2d_below=0.5,
    )
    ###########################################################################################################################################

    # selection de points DTM (max) sur une grille régulière
    # TODO: remplacer par GridDecimation une fois le correctif mergé dans PDAL
    pipeline |= pdal.Filter.grid_decimation_deprecated(
        resolution=0.5,
        output_dimension=dtm_dimension,
        output_type="max",
        where="(Classification==2 || PT_ON_SOL==0 && Classification==9)",
    )

    # selection de points DSM (max) sur une grille régulière
    # TODO: remplacer par GridDecimation une fois le correctif mergé dans PDAL
    pipeline |= pdal.Filter.grid_decimation_deprecated(
        resolution=0.5,
        output_dimension=dsm_dimension,
        output_type="max",
        where="(PT_ON_VEGET==0 && "
        + macro.build_condition("Classification", [6, 9, 17, 64])
        + f" || {dsm_dimension}==1)",
    )

    # assigne des points sol sélectionnés : les points proches de la végétation, des ponts, de l'eau, 67
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src=f"{dtm_dimension}==1",
        condition_ref=macro.build_condition("Classification", [4, 5, 6, 17, 67]),
        condition_out=f"{dsm_dimension}=0",
    )
    # Test proximité batiment
    pipeline = macro.add_radius_assign(
        pipeline,
        1.25,
        False,
        condition_src="Classification==2 && PT_VEG_DSM==0",
        condition_ref="Classification==6",
        condition_out="PT_ON_BUILDING=1",
    )
    # BUFFER INVERSE Se mettre
    pipeline = macro.add_radius_assign(
        pipeline,
        1,
        False,
        condition_src=f"Classification==2 && {dsm_dimension}==0 && PT_ON_BUILDING==1 && {dtm_dimension}==1",
        condition_ref="Classification==2 && PT_ON_BUILDING==0 && PT_VEG_DSM==0",
        condition_out=f"{dsm_dimension}=1",
    )
    # 3 - gestion des ponts
    #     bouche trou : on filtre les points au milieu du pont en les mettant à PT_ON_BRIDGE=1
    pipeline = macro.add_radius_assign(
        pipeline,
        1.5,
        False,
        condition_src=macro.build_condition("Classification", [2, 3, 4, 5, 6, 9, 67]),
        condition_ref="Classification==17",
        condition_out="PT_ON_BRIDGE=1",
        max2d_above=0,  # ne pas  prendre les points qui sont au dessus des points pont (condition_ref)
        max2d_below=-1,  # prendre tous les points qui sont en dessous des points pont (condition_ref)
    )
    pipeline = macro.add_radius_assign(
        pipeline,
        1.25,
        False,
        condition_src="PT_ON_BRIDGE==1",
        condition_ref="PT_ON_BRIDGE==0 && "
        + macro.build_condition("Classification", [2, 3, 4, 5, 6, 9, 67]),
        condition_out="PT_ON_BRIDGE=0",
        max2d_above=0.5,  # ne pas  prendre les points qui sont au dessus des points pont (condition_ref)
        max2d_below=0.5,  # prendre tous les points qui sont en dessous des points pont (condition_ref)
    )
    pipeline |= pdal.Filter.assign(value=[f"{dsm_dimension}=0 WHERE PT_ON_BRIDGE==1"])

    # 4 - point pour DTM servent au DSM également
    # HOMOGENEISER L UTILISATION DE PT_VEG_DSM POUR LES POINT SOL SOUS VEGET AVEC PT_ON_VEGET
    pipeline |= pdal.Filter.assign(
        value=[
            f"{dsm_dimension}=1 WHERE ({dtm_dimension}==1 && PT_VEG_DSM==0 && PT_ON_BRIDGE==0 && PT_ON_BUILDING==0 && PT_ON_VEGET==0)"
        ]
    )

    # 5 - Ajout de la classe 66 pts virtuels dans DTM et DSM
    pipeline |= pdal.Filter.assign(value=[f"{dtm_dimension}=1 WHERE (Classification==66)"])

    # 6 - export du nuage et des DSM
    pipeline |= pdal.Writer.las(extra_dims="all", forward="all", filename=output_las)

    return pipeline, temporary_dimensions


def mark_points_to_use_for_digital_models_with_new_dimension(
    input_las,
    output_las,
    dsm_dimension,
    dtm_dimension,
    output_dsm,
    output_dtm,
    keep_temporary_dimensions=False,
):
    with tempfile.NamedTemporaryFile(suffix="_with_temporary_dims.las", dir=".") as tmp_las:
        pipeline, temporary_dimensions = define_marking_pipeline(
            input_las,
            tmp_las.name,
            dsm_dimension,
            dtm_dimension,
        )

        if output_dtm:
            pipeline |= pdal.Writer.gdal(
                gdaldriver="GTiff",
                output_type="max",
                resolution=0.5,
                filename=output_dtm,
                where=f"{dtm_dimension}==1",
            )

        if output_dsm:
            pipeline |= pdal.Writer.gdal(
                gdaldriver="GTiff",
                output_type="max",
                resolution=0.5,
                filename=output_dsm,
                where=f"{dsm_dimension}==1",
            )

        pipeline.execute()

        if keep_temporary_dimensions:
            shutil.copy(tmp_las.name, output_las)
        else:
            remove_dimensions_from_las(
                tmp_las.name,
                temporary_dimensions + ["SRC_DOMAIN", "REF_DOMAIN", "radius_search"],
                output_las,
            )


def main(
    input_las,
    output_las,
    dsm_dimension,
    dtm_dimension,
    output_dsm,
    output_dtm,
    keep_temporary_dimensions=False,
    skip_buffer=False,
    buffer_width=25,
    spatial_ref="EPSG:2154",
    tile_width=1000,
    tile_coord_scale=1000,
):
    if skip_buffer:
        mark_points_to_use_for_digital_models_with_new_dimension(
            input_las,
            output_las,
            dsm_dimension,
            dtm_dimension,
            output_dsm,
            output_dtm,
            keep_temporary_dimensions,
        )
    else:
        mark_with_buffer = run_on_buffered_las(
            buffer_width, spatial_ref, tile_width, tile_coord_scale
        )(mark_points_to_use_for_digital_models_with_new_dimension)

        mark_with_buffer(
            input_las,
            output_las,
            dsm_dimension,
            dtm_dimension,
            output_dsm,
            output_dtm,
            keep_temporary_dimensions,
        )


if __name__ == "__main__":
    args = parse_args()
    main(**vars(args))
